/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/* ADT of doubly linked list using external storage model.
 * see http://en.wikipedia.org/wiki/Linked_list for detail.
 */
#if defined(ADT_DEBUG) || defined(ADT_DLLIST_DEBUG)
#include <stdio.h>
#endif
#include <stdlib.h>
#include <stdarg.h>
#include "dllist.h"

/* Init DLL structure to zero. */
void dll_init(dll_t *dll)
{
	dll->linklen = 0, dll->head = dll->tail = NULL;
}

/* Duplicate another DLL, pointing to the same external storage. */
bool dll_dup(dll_t *dst, const dll_t *src)
{
	dll_node_t *p = src->head, *q, *t;
#if defined(ADT_DEBUG)
	unsigned i = 1;
#endif
	/* sanity check. */
	if(!p) {
#if defined(ADT_DEBUG)
		printf("dll_dup(): src->head=NULL, sanity check failed!\n");
#endif
		return false;
	}

#if defined(ADT_DEBUG)
	printf("dll_dup(): assign dst head node.\n");
#endif
	/* allocate dst head node. */
	t = dst->head = malloc(sizeof(dll_node_t));
	if(!t) {
#if defined(ADT_DEBUG)
		printf("dll_dup(): malloc() failed!\n");
#endif
		return false;
	}
	t->data = p->data, t->prev = NULL, p = p->next;

	/* copy src nodes to dst DLL one by one. */
	while(p) {
		q = malloc(sizeof(dll_node_t));
		if(!q) {
#if defined(ADT_DEBUG)
			printf("dll_dup(): malloc() failed!\n");
#endif
			t->next = NULL;
			dll_destroy(dst);
			return false;
		}
		q->data = p->data, p = p->next;
		t->next = q, q->prev = t, t = q;
#if defined(ADT_DEBUG)
		printf("dll_dup(): copy src node to dst DLL %u times...\n", i++);
#endif
	}

#if defined(ADT_DEBUG)
	printf("dll_dup(): assign dst tail node & linklen.\n");
#endif
	/* assign dst tail node and linklen. */
	dst->tail = t, t->next = NULL;
	dst->linklen = src->linklen;
	return true;
}

/* Peek specific node's element content,
 * return the pointer on valid offset, otherwise NULL. */
const void *dll_node_peek(dll_t *dll, size_t offset)
{
	size_t i;
	dll_node_t *r;
	
#ifdef ADT_DEBUG
	printf("dll_node_peek(): visiting elem...\n");
#endif
	/* sanity check. */
	if(offset >= dll->linklen) {
#ifdef ADT_DEBUG
		printf("dll_node_peek(): offset > dll->linklen, sanity check failed!\n");
#endif
		return NULL;
	}

	/* positioning... */
	i = 0, r = dll->head;
	while(i < offset) {
#ifdef ADT_DEBUG
		printf("dll_node_peek(): i=%u < offset=%u, positioning need.\n", i, offset);
#endif
		++i, r = r->next;
	}
	return r->data;
}

/* Delete DLL's current head node. */
void dll_delhead(dll_t *dll)
{
	dll_node_t *t = dll->head;
	
	if(!t) {
#if defined(ADT_DEBUG)
		printf("dll_delhead(): dll->head=NULL, sanity check failed!\n");
#endif
		return;
	}
	dll->head = t->next;
	if(--dll->linklen) {
		dll->head->prev = NULL;
	} else { /* DLL contains only one node. */
#if defined(ADT_DEBUG)
		printf("dll_delhead(): this is the last node.\n");
#endif
		dll->tail = NULL;
	}
#if defined(ADT_DEBUG)
	printf("dll_delhead(): free(%p)...\n", t);
#endif
	free(t);
}

/* Delete DLL's current tail node. */
void dll_deltail(dll_t *dll)
{
	dll_node_t *q = dll->tail; 

	/* sanity check. */
	if(!q) {
#if defined(ADT_DEBUG)
		printf("dll_deltail(): dll->tail=NULL, sanity check failed!\n");
#endif
		return;
	}
	
	if(!--dll->linklen) { /* DLL contains only one node. */
#if defined(ADT_DEBUG)
		printf("dll_deltail(): this is the last node.\n");
#endif
		dll->head = dll->tail = NULL;
	} else { 
		dll->tail = q->prev, dll->tail->next = NULL;
	}
#if defined(ADT_DEBUG)
	printf("dll_deltail(): free(%p)...\n", q);
#endif
	free(q);
}

/* Scan link and update matching node(s),(test() returns true)
 * return number of updated node(s). */
size_t dll_node_cupdt(dll_t *dll, bool (*test)(const void *), const void *data)
{
	size_t cnt = 0;
	dll_node_t *q = dll->head;
	
	for(; q; q = q->next) {
		if( (*test)(q->data)) {
#if defined(ADT_DEBUG)
			printf("dll_node_cupdt(): matching, updating...\n");
#endif
			q->data = data, ++cnt;
		}
	}
#if defined(ADT_DEBUG)
	printf("dll_node_cupdt(): updated %u times...\n", cnt);
#endif
	return cnt;
}
	
/* Add a new head node. */
bool dll_addhead(dll_t *dll, const void *data)
{
	dll_node_t *t = dll->head, *r;
	
	/* allocate new head node. */
	r = malloc(sizeof(dll_node_t));
	if(!r) {
#if defined(ADT_DEBUG)
		printf("dll_addhead(): malloc() failed!\n");
#endif
		return false;
	}
	dll->head = r, r->data = data;
	r->next = t, r->prev = NULL;
	
	/* assign new head node. */	
	dll->linklen++ ? t->prev = r : (dll->tail = r);
#if defined(ADT_DEBUG)
	printf("dll_addhead(): linklen=%u\n", dll->linklen);
#endif
	return true;
}

/* Add a new tail node (append). */
bool dll_addtail(dll_t *dll, const void *data)
{
	dll_node_t *t = dll->tail, *r; 

	/* allocate new tail node. */
	r = malloc(sizeof(dll_node_t));
	if(!r) {
#if defined(ADT_DEBUG)
		printf("dll_addtail(): malloc() failed!\n");
#endif
		return false;
	}
	dll->tail = r, r->data = data;
	r->next = NULL, r->prev = t;

	/* assign new tail node. */	
	dll->linklen++ ? t->next = r : (dll->head = r);
#if defined(ADT_DEBUG)
	printf("dll_addtail(): linklen=%u\n", dll->linklen);
#endif
	return true;
}

/* Conditionally (callback returns true) delete DLL node(s),
 * return number of deleted node(s). */
size_t dll_node_cdel(dll_t *dll, bool (*test)(const void *))
{
	size_t cnt = 0;
	dll_node_t *q = dll->head, *t;

	/* head node matches. */
	while(q && (*test)(q->data)) {
#if defined(ADT_DEBUG)
		printf("dll_node_cdel(): head node matches...\n");
#endif
		dll->head = q->next, dll->head->prev = NULL;
		free(q);
		q = dll->head, ++cnt;
	}

#if defined(ADT_DEBUG)
	printf("dll_node_cdel(): head node doesn't match.\n");
#endif
	/* head node doesn't match. */
	while(q) {
		t = q->next;
		if(!t) { /* this is the tail node. */
#if defined(ADT_DEBUG)
			printf("dll_node_cdel(): this is the tail node.\n");
#endif
			break;
		} else if( (*test)(t->data)) { /* next node matches. */
#if defined(ADT_DEBUG)
			printf("dll_node_cdel(): node matches...\n");
#endif
			q->next = t->next; 
			if(t->next)
				t->next->prev = t->prev;
			free(t), ++cnt;
		} else { /* next node doesn't match. */
#if defined(ADT_DEBUG)
			printf("dll_node_cdel(): node doesn't match.\n");
#endif
			q = t;
		}
	}

	dll->tail = q; 
	dll->linklen -= cnt;
#if defined(ADT_DEBUG)
	printf("dll_node_cdel(): deleted %u times...\n", cnt);
#endif
	return cnt;
}

/* Insert a node right before matching node (callback returns true),
 * return true on success, otherwise false. */
bool dll_node_bins(dll_t *dll, bool (*test)(const void *), const void *data)
{
	dll_node_t *q = dll->head, *t, *r; 

	for(; q; q = q->next) {
		if( (*test)(q->data)) {
			/* allocate new node. */
			r = malloc(sizeof(dll_node_t));
			if(!r) {
#if defined(ADT_DEBUG)
				printf("dll_node_bins(): malloc() failed!\n");
#endif
				return false;
			}
			t = q->prev, r->data = data;

			/* assign new node. */
			r->next = q, q->prev = r, r->prev = t;
			t ? t->next = r : (dll->head = r);
			++dll->linklen;
#if defined(ADT_DEBUG)
			printf("dll_node_bins(): found and assigned.\n");
#endif
			return true;
		}
	}
#if defined(ADT_DEBUG)
	printf("dll_node_bins(): not found.\n");
#endif
	return false;
}

/* Insert a node right after matching node (callback returns true),
 * return true on success, otherwise false. */
bool dll_node_ains(dll_t *dll, bool (*test)(const void *), const void *data)
{
	dll_node_t *q = dll->head, *t, *r; 

	for(; q; q = q->next) {
		if( (*test)(q->data)) {
			/* allocate new node. */
			r = malloc(sizeof(dll_node_t));
			if(!r) {
#if defined(ADT_DEBUG)
				printf("dll_node_ains(): malloc() failed!\n");
#endif
				return false;
			}
			t = q->next, r->data = data;

			/* assign new node. */
			q->next = r, r->prev = q, r->next = t;
			t ? t->prev = r : (dll->tail = r);
			++dll->linklen;
#if defined(ADT_DEBUG)
			printf("dll_node_ains(): found and assigned.\n");
#endif
			return true;
		}
	}
#if defined(ADT_DEBUG)
	printf("dll_node_ains(): not found.\n");
#endif
	return false;
}

/* Return number of node(s) DLL contains. */
size_t dll_getlen(dll_t *dll)
{
	return dll->linklen;
}

static __inline dll_node_t *dll_getmaxormin(dll_node_t *p, size_t len, int (*cmp)(const void *, const void *))
{
	dll_node_t *t, *r; 
	size_t i;
#if defined(ADT_DEBUG)
	printf("dll_getmaxormin(): len=%u\n", len);
#endif
	for(r=t=p,i=1; i<len; ++i,r=r->next) 
		if( (*cmp)(t->data, r->next->data) > 0 ) {
			t = r->next;
		}
#if defined(ADT_DEBUG)
	printf("dll_getmaxormin(): returning t=%p\n", t);
#endif
	return t;
}

/* Sort DLL content by exchanging const pointers to external storage. */
void dll_sort(dll_t *dll, int (*cmp)(const void *, const void *))
{
	dll_node_t *q = dll->head, *r;
	const void *t;
	size_t len = dll->linklen;

#if defined(ADT_DEBUG)
	printf("dll_sort(): selection sorting link..., dll->linklen=%u\n", len);
#endif
	
	if(!q) {
#if defined(ADT_DEBUG)
		printf("dll_sort(): dll->head=NULL, sanity check failed!\n");
#endif
		return;
	}

	/* selection sort. */
	for(; q->next; q = q->next) {
		/* getting node with max/min node content. */
		r = dll_getmaxormin(q, --len, cmp);

		/* exchange two pointers. */
		if(r != q) {
			t = r->data, r->data = q->data, q->data = t;
		}
	}
}

static __inline dll_t merge2link(dll_t *p, dll_t *q, int (*cmp)(const void *, const void *))
{
	dll_node_t *t, *a = p->head, *b = q->head;
	dll_t h;

	if(!b) {
#if defined(ADT_DEBUG)
		printf("merge2link(): sanity check failed, q=NULL\n");
#endif
		h = *p;
		goto done;
	}
	if(!a) {
#if defined(ADT_DEBUG)
		printf("merge2link(): sanity check failed, p=NULL\n");
#endif
		h = *q;
		goto done;
	}

#if defined(ADT_DEBUG)
	printf("merge2link(): assigning new head...");
#endif
	/* assign new head, length for incoming new DLL. */
	if( (*cmp)(&a->data, &b->data) < 0 )
		h = *p, t = a, a = a->next, h.linklen += q->linklen;
	else
		h = *q, t = b, b = b->next, h.linklen += p->linklen;

#if defined(ADT_DEBUG)
	printf("done.\nmerge2link(): merge sorting...");
#endif
	/* merge sort. */
	while(a && b) {
		if( (*cmp)(&a->data, &b->data) < 0 )
			t = t->next = a, a = a->next;
		else
			t = t->next = b, b = b->next;
	}

#if defined(ADT_DEBUG)
	printf("done.\nmerge2link(): handling remaining part...");
#endif
	/* attach remaining part. */
	if(a) 
		t->next = a, h.tail = p->tail;
	else
		t->next = b, h.tail = q->tail;

#if defined(ADT_DEBUG)
	printf("done.\n");
#endif
	/* cleanup input's data structure. */
done:
	dll_init(p);
	dll_init(q);
	return h;
}

/* Merge asc/desc sorted links (all in same order) into a single sorted dst, 
 * various dll_t * parameters and a terminating NULL is expected. 
 * original ones will be destroyed automatically. */
void dll_merge(dll_t *dst, int (*cmp)(const void *, const void *), ...)
{
	va_list p;
	dll_t *t;

#if defined(ADT_DEBUG)
	printf("dll_merge(): merging begins...\n");
#endif
	dll_init(dst);
	va_start(p, cmp);
	for(;;) {
		t = va_arg(p, dll_t *);
		if(!t) {
#if defined(ADT_DEBUG)
			printf("dll_merge(): done with arg parsing.\n");
#endif
			break;
		}
		*dst = merge2link(dst, t, cmp);
	}
	va_end(p);
}

/* Traverse DLL via visit() from head to tail if flag is true,
 * else from tail to head. */
void dll_trav(dll_t *dll, bool flag, void (*visit)(const void *))
{
	dll_node_t *q;
	size_t i=0, l = dll->linklen;

	if(flag) {
#if defined(ADT_DEBUG)
		printf("dll_trav(): visiting link(length=%u) from head to tail...\n", dll->linklen);
#endif
		for(q=dll->head; i<l; ++i, q=q->next)
			(*visit)(q->data);	
	} else {
#if defined(ADT_DEBUG)
		printf("dll_trav(): visiting link(length=%u) from tail to head...\n", dll->linklen);
#endif
		for(q=dll->tail; i<l; ++i, q=q->prev)
			(*visit)(q->data);	
	}
}

/* Free DLL allocated memory. */
void dll_destroy(dll_t *dll)
{
	dll_node_t *q = dll->head, *t;

	if(!q) {
#if defined(ADT_DEBUG)
		printf("dll_destroy(): sanity check failed, empty DLL already!\n");
#endif
		return;
	}

#if defined(ADT_DEBUG)
	printf("dll_destroy(): linklen=%u, do it from head to tail\n", dll->linklen);
#endif
	while(q) {
#if defined(ADT_DEBUG)
		printf("dll_destroy(): freeing(%p)...\n", q);
#endif
		t = q->next;
		free(q);
		q = t;
	}
	dll_init(dll);
}

#if defined(ADT_DLLIST_DEBUG)
#define DLL_MAX_NODE 8
static bool uptest(dll_t *dll)
{
	printf("uptest(): dll->data = %d\t", dll->data);
	return dll->data > 20 ? true : false;
}
static bool deltest(dll_t *dll)
{
	printf("deltest(): dll->data = %d\t", dll->data);
	return dll->data > 1000 ? true : false;
}
static bool ainstest(dll_t *dll)
{
	printf("ainstest(): condition is: dll->data > 2000, dll->data = %d\n", dll->data);
	return dll->data > 2000 ? true : false;
}

static int sortcmp(dll_t *dll, const void *data)
{
	return q->data - dll->data;
}
static int sortcmp2(dll_t *dll, const void *data)
{
	return dll->data - q->data;
}
static void visit(dll_t *dll)
{
	printf("\tvisit(): data=%d\n", dll->data);
}
int main(void)
{
	dll_t list1, list2, list3, list4, *p = &list1, *n = &list2, *m=&list4;
	dll_elem_t elem, *q = &elem;
	int i;

	dll_init(p);
	for(i=0; i<DLL_MAX_NODE; i++) {
		q->data = i*i;
		dll_addtail(p, q);	
	}

	for(i=0; i<DLL_MAX_NODE*2; i++) {
		if((q=dll_node_peek(p,i)))
			printf("\tindex=%d, data=%d\n",i, q->data);
	}
	for(i=DLL_MAX_NODE; i>=0; i--) {
		if((q=dll_node_peek(p,i)))
			printf("\tindex=%d, data=%d\n",i, q->data);
	}

	dll_dup(n, p);
	for(i=0; i<DLL_MAX_NODE; i++) {
		if((q=dll_node_peek(p,i)))
			printf("\tindex=%d, data=%d\n",i, q->data);
	}
	for(i=DLL_MAX_NODE-1; i>=0; i--) {
		if((q=dll_node_peek(p,i)))
			printf("\tindex=%d, data=%d\n",i, q->data);
	}

	for(i=0; i<DLL_MAX_NODE*2; i++) {
		dll_delhead(n);
	}
	dll_destroy(n);

	dll_dup(n, p);
	for(i=0; i<DLL_MAX_NODE*2; i++) {
		dll_deltail(n);
	}

	dll_dup(n, p);
	q->data = 999;
	dll_node_cupdt(n, uptest, q);
	dll_trav(n, true, visit);

	for(i=0; i<5; i++) {
		q->data = i+1000;
		dll_addhead(n, q);
	}
	dll_trav(n, true, visit);

	for(i=0; i<5; i++) {
		q->data = i+2000;
		dll_addtail(n, q);
	}
	dll_trav(n, false, visit);

	dll_node_cdel(p, deltest);
	dll_trav(n, true, visit);
	dll_node_cdel(n, deltest);
	dll_trav(n, false, visit);
	
	q->data = 3000;
	dll_addtail(n, q);
	q->data = 9999;
	dll_node_ains(n, ainstest, q);
	dll_trav(n, true, visit);

	q->data = 8888;
	dll_node_bins(n, uptest, q);
	dll_trav(n, false, visit);
#if 0
	dll_tort(n, sortcmp);
	dll_tort(p, sortcmp);
	dll_dup(m, n);
	dll_merge(&list3, sortcmp, m, n, p, NULL);
	dll_trav(&list3, true, visit);
	dll_trav(m, true, visit);
	dll_trav(n, true, visit);
	dll_trav(p, true, visit);
	dll_destroy(&list3);
#endif
#if 1
	dll_tort(n, sortcmp2);
	printf("travelling throughout n...");
	dll_trav(n, true, visit);
	dll_tort(p, sortcmp2);
	printf("travelling throughout p...");
	dll_trav(p, true, visit);
	dll_dup(m, n);
	dll_merge(&list3, sortcmp2, m, n, p, NULL);
	dll_trav(&list3, true, visit);
	dll_trav(m, true, visit);
	dll_trav(n, true, visit);
	dll_trav(p, true, visit);
	dll_destroy(&list3);
#endif
#if 0
	dll_destroy(n);
	dll_destroy(p);
#endif
	return 0;
}
#endif
