/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/* ADT of chained hash table operations.
 */
#if defined(ADT_DEBUG) || defined(ADT_HASH_DEBUG)
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>
#include "hash.h"

/* Create and init hash table structure. */
bool hash_create(hash_t *p, size_t maxbucks, hash_func hashf)
{
	p->memstart = calloc(maxbucks+1, sizeof(hash_buck_t));
	if(!p->memstart)
		return false;
	p->maxbucks = maxbucks;
	p->hashfunc = hashf;
	return true;
}

/* Insert an item to the table. */
bool hash_insert(hash_t *p, const char *key, const void *data)
{
	int index = (*p->hashfunc)(key);
	hash_node_t *t = (p->memstart+index)->head, *r;

	/* allocate new node. */
	r = malloc(sizeof(hash_node_t));
	if(!r) {
		return false;
	}
	r->data = data, r->key = key;

	if(!t) {
		t = r, r->next = NULL;
	} else {
		r->next = t, t = r;
	}
	return true;
}

/* Lookup an item within the table. */
const void *hash_lookup(hash_t *p, const char *k)
{
	int index = (*p->hashfunc)(k);
	hash_node_t *r = (p->memstart+index)->head;
	while(r && memcmp(&r->key, k, sizeof(char))) {
		r = r->next;
	}
	return r ? &r->data : NULL;
}

/* Remove one or more items from the table. */
void hash_remove(hash_t *p, const char *k)
{
	int index = (*p->hashfunc)(k);
	hash_node_t *q = (p->memstart+index)->head, *r =q, *t;

	/* head node matches. */
	while(q && !memcmp(&q->key, k, sizeof(char))) {
		r = q->next;
		free(q);
		q = r;
	}

	/* head node doesn't match. */
	while(q) {
		t = q->next;
		if(!t) { /* this is the tail node. */
			break;
		} else if( !memcmp(&t->key, k, sizeof(char))) { /* next node matches. */
			q->next = t->next;
			free(t);
		} else { /* next node doesn't match. */
			q = t;
		}
	}
	(p->memstart+index)->head = q;
}

/* Destroy and free everything. */
void hash_destroy(hash_t *p)
{
	size_t i = 0;
	hash_node_t *node, *t;

	while(i <= p->maxbucks) {
		node = (p->memstart+i++)->head;
		while(node) {
			t = node->next;
			free(node);
			node = t;
		}
	}
	free(p->memstart);
}

#ifdef ADT_HASH_DEBUG
/* Hash n bytes memory and return hash value. */
static unsigned hash_hash(const void *src, size_t n, unsigned divisor)
{
	int d = 0, i = 0;
	div_t t = div(n, sizeof(int));
	const char *p = (const char *)src;

	while(i++ < t.rem) {
		d += *p++;
	}
	for(i=0; i<t.quot; ++i) {
		d += *(const int *)p++;
	}
	return (unsigned)d % divisor;
}

int main(void)
{
	return 0;
}
#endif
