/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/* ADT of stack (LIFO), in 3 forms.
 * DYNAMIC: slowest but flexible, handy when elem structure is really big.
 * STATIC: fastest but must know maximum size/len of stack.
 * BALANCED: general purpose stack, balanced performance/memory usage.
 * see http://en.wikipedia.org/wiki/Stack_(data_structure) for detail.
 */
#include <stdlib.h>
#include "stack.h"

/* Init stack ADT, call this before any following operations. */
void stack_init(stack_t *p)
{
#ifdef ADT_STACK_BALANCED
	p->top = p->base = NULL; 
	p->memunits = 0;
#endif
#ifdef ADT_STACK_STATIC
	p->top = p->base = p->e;
#endif
#ifdef ADT_STACK_DYNAMIC
	sll_init(p);
#endif
}

/* Push a node into stack. return false on malloc() failure, else true. */
bool stack_push(stack_t *p, const void *data)
{
#ifdef ADT_STACK_BALANCED
	if( (size_t)(p->top - p->base) == p->memunits) {
		void *t = realloc(p->base, (p->memunits + STACK_INCREMENT) * sizeof(stack_node_t));
		if(!t) {
			return false;		
		}
		p->base = t;
		p->top = p->base + p->memunits;
		p->memunits += STACK_INCREMENT;
	}
	p->top->data = data, ++p->top;
	return true;
#endif
#ifdef ADT_STACK_STATIC
	p->top->data = data, ++p->top;
	return true;
#endif
#ifdef ADT_STACK_DYNAMIC
	return sll_addhead(p, data);
#endif
}

/* Popup top node from stack. 
 * return external storage pointer or NULL on empty stack. */
const void *stack_pop(stack_t *p)
{
#ifdef ADT_STACK_BALANCED
	if(p->top == p->base)
		return NULL;
	return (--p->top)->data;
#endif
#ifdef ADT_STACK_STATIC
	return (--p->top)->data;
#endif
#ifdef ADT_STACK_DYNAMIC
	const void *t ;
	if(!p->head)
		return NULL;
	t = p->head->data;
	sll_delhead(p);
	return t;
#endif
}

/* Whether stack is empty(no node) or not. */
bool stack_isempty(stack_t *p)
{
#ifdef ADT_STACK_BALANCED
	return p->top != p->base ? false : true;
#endif
#ifdef ADT_STACK_STATIC
	return p->top != p->base ? false : true;
#endif
#ifdef ADT_STACK_DYNAMIC
	return sll_getlen(p) ? true : false;
#endif
}

/* Clear stack content, making it empty/reusable. */
void stack_clear(stack_t *p)
{
#ifdef ADT_STACK_BALANCED
	p->top = p->base;
#endif
#ifdef ADT_STACK_STATIC
	p->top = p->base;
#endif
#ifdef ADT_STACK_DYNAMIC
	sll_destroy(p);
#endif
}

/* Peek specific node's element content,
 * return the pointer on valid offset, otherwise NULL. */
const void *stack_peek(stack_t *p, size_t t)
{
#ifdef ADT_STACK_BALANCED
	if(p->top == p->base)
		return NULL;
	return (p->base + t);
#endif
#ifdef ADT_STACK_STATIC
	return (p->base + t);
#endif
#ifdef ADT_STACK_DYNAMIC
	return sll_node_peek(p, t);
#endif
}

/* Get number of node(s) stack contains. */
size_t stack_getlen(stack_t *p)
{
#ifdef ADT_STACK_BALANCED
	return p->top - p->base;
#endif
#ifdef ADT_STACK_STATIC
	return p->top - p->base;
#endif
#ifdef ADT_STACK_DYNAMIC
	return sll_getlen(p);
#endif
}

/* Free allocated stack memory. */
void stack_destroy(stack_t *p)
{
#ifdef ADT_STACK_BALANCED
	free(p->base);
#endif
#ifdef ADT_STACK_STATIC
#endif
#ifdef ADT_STACK_DYNAMIC
	sll_destroy(p);
#endif
}
